import os

import typer
from rich.console import Console
from rich.panel import Panel

from .__version__ import __version__
from .audio import Audio
from .fav import Fav
from .season import Season
from .series import Series
from .user import User
from .util import AudioQuality, AudioQualityEnums, Directory, Path

app = typer.Typer()

console = Console()


@app.callback(invoke_without_command=True)
def callback(version: bool = False):
    """
    Entry for public options
    """

    if version:
        typer.echo(f"ppatch, version {__version__}")
        raise typer.Exit()


@app.command()
def bv(
    bv: str,
    directory: Directory = None,
    audio_quality: AudioQuality = AudioQualityEnums.k192.value,
):
    audio = Audio(bv, audio_quality)
    audio.download()


@app.command()
def uid(
    uid: str,
    directory: Directory = None,
    audio_quality: AudioQuality = AudioQualityEnums.k192.value,
):
    user = User(uid)

    for video in user.videos:
        bv = video["bvid"]
        audio = Audio(bv, audio_quality)
        audio.download()

    console.print(
        Panel(
            f"已下载 UID {uid} 的所有音频",
            subtitle="下载完成",
            title="UID 下载",
            style="bold green",
            expand=False,
        )
    )


@app.command()
def fav(
    media_id: str,
    cookies_path: str = Path,
    directory: Directory = None,
    audio_quality: AudioQuality = AudioQualityEnums.k192.value,
):
    with open(cookies_path, "r") as f:
        cookies = f.read()

    fav = Fav(media_id, cookies)

    for bvid in fav.id_list:
        audio = Audio(bvid, audio_quality)
        audio.download()

    console.print(
        Panel(
            f"已下载 收藏夹 {media_id} 的所有音频",
            subtitle="下载完成",
            title="收藏夹下载",
            style="bold green",
            expand=False,
        )
    )


@app.command()
def season(
    uid: str,
    sid: str,
    directory: Directory = None,
    audio_quality: AudioQuality = AudioQualityEnums.k192.value,
):
    sea = Season(uid, sid)
    audio_generator = sea.get_videos()
    if not audio_generator:
        raise typer.Exit(1)

    if not os.path.isdir(sea.name):
        os.makedirs(sea.name)
    os.chdir(sea.name)

    for audios in audio_generator:
        for id in audios:
            audio = Audio(id, audio_quality)
            audio.download()
    console.print(
        Panel(
            f"已下载 合集 {sid} 的所有音频",
            subtitle="下载完成",
            title="合集下载",
            style="bold green",
            expand=False,
        )
    )


@app.command()
def series(
    uid: str,
    sid: str,
    directory: Directory = None,
    audio_quality: AudioQuality = AudioQualityEnums.k64,
):
    """Download bilibili video series

    The api of series lacks the series name, executing
    this command will not create a folder for the series
    """
    ser = Series(uid, sid)
    audio_generator = ser.get_videos()
    if not audio_generator:
        raise typer.Exit(1)

    for audios in audio_generator:
        for id in audios:
            audio = Audio(id, audio_quality)
            audio.download()
    console.print(
        Panel(
            f"已下载 列表 {sid} 的所有音频",
            subtitle="下载完成",
            title="列表下载",
            style="bold green",
            expand=False,
        )
    )



import os
import time

import requests
from rich.console import Console
from rich.panel import Panel
from rich.progress import BarColumn, DownloadColumn, Progress, TransferSpeedColumn

from .util import AudioQualityEnums, get_signed_params

console = Console()


class Audio:
    bvid = ""
    title = ""
    playUrl = "http://api.bilibili.com/x/player/wbi/playurl"
    part_list = []

    headers = {}

    def __init__(self, bvid: str, audio_quality: AudioQualityEnums) -> None:
        if bvid is None:
            raise ValueError("bvid is None")

        self.bvid = bvid
        self.headers = {
            "authority": "api.bilibili.com",
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
            "accept-language": "zh-CN,zh;q=0.9",
            "cache-control": "no-cache",
            "dnt": "1",
            "pragma": "no-cache",
            "sec-ch-ua": '"Not A(Brand";v="99", "Google Chrome";v="121", "Chromium";v="121"',
            "sec-ch-ua-mobile": "?0",
            "sec-ch-ua-platform": '"macOS"',
            "sec-fetch-dest": "document",
            "sec-fetch-mode": "navigate",
            "sec-fetch-site": "none",
            "sec-fetch-user": "?1",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
            "Referer": "https://www.bilibili.com/video/{bvid}".format(bvid=self.bvid),
        }

        self.audio_quality = audio_quality.quality_id

        # 获取cid和title
        if len(bvid) == 12:
            # BV号
            self.__get_cid_title(bvid)
        else:
            # AnotherV号
            self.__get_cid_title(bvid[:12])

    def download(self):
        start_time = time.time()
        try:
            for cid, part in zip(self.cid_list, self.part_list):
                if len(self.part_list) > 1:
                    file_path = f"{self.title}-{part}.mp3"
                else:
                    file_path = f"{self.title}.mp3"

                if len(file_path) > 255:
                    file_path = file_path[:255]

                # 如果文件已存在，则跳过下载
                if os.path.exists(file_path):
                    console.print(
                        Panel(
                            f"{self.title} 已存在，跳过下载",
                            style="yellow",
                            expand=False,
                        )
                    )
                    return

                params = get_signed_params(
                    {
                        "fnval": 16,
                        "bvid": self.bvid,
                        "cid": cid,
                    }
                )
                res = requests.get(
                    self.playUrl, params=params, headers=self.headers, timeout=60
                )

                json = res.json()

                if json["data"] is None:
                    console.print(
                        Panel(
                            f"[bold red]数据字段无效[/bold red]\n"
                            f"URL: {self.playUrl}\n"
                            f"参数: {params}",
                            title="错误",
                            expand=False,
                        )
                    )
                    return

                audio = json["data"]["dash"]["audio"]
                if not audio:
                    console.print(
                        Panel(
                            f"[bold red]音频字段为空[/bold red]\n"
                            f"URL: {self.playUrl}\n"
                            f"参数: {params}",
                            title="错误",
                            expand=False,
                        )
                    )
                    return

                base_url = None
                for au in audio:
                    if au["id"] == self.audio_quality:
                        base_url = au["baseUrl"]

                # no audio url corresponding to current audio quality
                if base_url is None:
                    base_url = audio[0]["baseUrl"]

                response = requests.get(url=base_url, headers=self.headers, stream=True)

                total_size = int(response.headers.get("content-length", 0))

                with Progress(
                    "[progress.description]{task.description}",
                    BarColumn(),
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    "•",
                    DownloadColumn(),
                    "•",
                    TransferSpeedColumn(),
                    console=console,
                ) as progress:
                    task = progress.add_task(
                        f"[cyan]下载 {self.title}", total=total_size
                    )

                    with open(file_path, "wb") as f:
                        for chunk in response.iter_content(chunk_size=8192):
                            if chunk:
                                f.write(chunk)
                                progress.update(task, advance=len(chunk))

                    # 添加下载完成的提示
                    end_time = time.time()
                    download_time = round(end_time - start_time, 2)
                    console.print(
                        Panel(
                            f"[bold green]下载完成！[/bold green]用时 {download_time} 秒",
                            expand=False,
                        )
                    )

        except Exception as e:
            console.print(
                Panel(
                    f"[bold red]下载失败[/bold red]\n Code: {res.status_code} 错误: {str(e)}",
                    title="异常",
                    expand=False,
                )
            )
            raise e

    def __get_cid_title(self, bvid: str):
        url = "https://api.bilibili.com/x/web-interface/view"
        params = {"bvid": bvid}

        try:
            response = requests.get(
                url=url,
                params=params,
                headers=self.headers,
            )
            data = response.json().get("data")
            self.title = self.__title_process(data.get("title"))

            self.cid_list = [str(page.get("cid")) for page in data.get("pages", [])]
            self.part_list = [
                self.__title_process(str(page.get("part")))
                for page in data.get("pages", [])
            ]

        except Exception as e:
            console.print(
                Panel(
                    f"[bold red]获取视频信息失败[/bold red]\n错误: {str(e)}",
                    title="异常",
                    expand=False,
                )
            )
            raise

    def __title_process(self, title: str):
        replaceList = ["?", "\\", "*", "|", "<", ">", ":", "/", " "]
        for ch in replaceList:
            title = title.replace(ch, "-")

        return title



"""download bilibili season archive, 视频合集下载"""

import typer

from .util import Retry, request

headers: dict[str, str] = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
    "Referer": "https://www.bilibili.com",
}


class Season:
    # api refer https://github.com/SocialSisterYi/bilibili-API-collect/blob/f9ee5c3b99335af6bef0d9d902101c565b3bea00/docs/video/collection.md
    season_url: str = (
        "https://api.bilibili.com/x/polymer/web-space/seasons_archives_list"
    )
    retry = 3

    def __init__(self, uid: str, sid: str, page_size=30) -> None:
        self.uid = uid
        self.season_id = sid
        self.page_size = page_size
        self.videos = []

    def get_videos(self):
        params = {
            "mid": self.uid,
            "season_id": self.season_id,
            "sort_reverse": False,
            "page_num": 1,
            "page_size": self.page_size,
        }

        @Retry(self.__response_succeed, self.__handle_error_response)
        def wrapped_request():
            """wrap request with retry"""
            return request(
                method="get", url=self.season_url, params=params, headers=headers
            ).json()

        res = wrapped_request()
        if res is None:
            return False

        self.total = res["data"]["meta"]["total"]
        self.name = res["data"]["meta"]["name"]

        def bvid_generator():
            max_pn = self.total // self.page_size
            for i in range(1, max_pn + 2):
                params["page_num"] = i
                res = wrapped_request()
                if res:
                    bvids = [d["bvid"] for d in res["data"]["archives"]]
                    # self.videos.extend(bvids)
                    yield bvids
                else:
                    typer.echo(
                        f"skip audios from {(i-1)* self.page_size} to {i * self.page_size}"
                    )

        return bvid_generator()

    def __handle_error_response(self, response):
        code = response.get("code", -404)
        if code == -404:
            typer.echo(f"Error: uid {self.uid} or sid {self.season_id} error.")
        elif code == -352:
            typer.echo(
                "Error: Authentication problem or too many requests, please try again later."
            )
        else:
            typer.echo("Error: Unknown problem.")

        typer.echo(f"code: {response['code']}")
        if response.get("message", None):
            typer.echo(f"msg: {response['message']}")

    def __response_succeed(self, response):
        return response.get("code", -404) == 0



import typer

from .util import get_w_webid, request


class User:
    uidUrl: str = "https://api.bilibili.com/x/space/wbi/arc/search"
    uid: str = ""
    videos: list = []

    def __init__(self, uid: str):
        self.uid = uid

        w_webid = get_w_webid(uid)

        params = {
            "mid": uid,
            "ps": 1,
            "tid": 0,
            "pn": 1,
            "order": "pubdate",
            "web_location": "333.999",
            "w_webid": w_webid,
        }

        response = request(
            method="get", url=self.uidUrl, params=params, wbi=True, dm=True
        )

        code = response.json().get("code", -400)

        if code != 0:
            typer.echo(f"Error: uid {uid} not found")
            raise typer.Exit(1)

        total = response.json()["data"]["page"]["count"]

        max_pn = total // 50
        surpus = total % 50

        # get all videos
        for i in range(1, max_pn + 2):
            ps = 50 if i != max_pn + 1 else surpus

            params = {
                "mid": uid,
                "ps": ps,
                "tid": 0,
                "pn": i,
                "order": "pubdate",
                "web_location": "333.999",
                "w_webid": w_webid,
            }

            response = request(
                method="get", url=self.uidUrl, params=params, wbi=True, dm=True
            )

            self.videos.extend(response.json()["data"]["list"]["vlist"])



import json

from .util import request


class Fav:
    media_id = ""
    cookies = {}
    fav_url = "https://api.bilibili.com/x/v3/fav/resource/ids"

    id_list = []

    def __init__(self, media_id: str, cookies_str: str) -> None:
        if cookies_str is None or cookies_str is None:
            raise ValueError("Error: Params are Required")

        try:
            self.cookies = json.loads(cookies_str)
        except Exception as e:
            raise ValueError("Error: Cookies is not a json string")

        # 获取 id list

        params = {
            "media_id": media_id.replace("ml", ""),
            "platform": "web",
        }

        try:
            res = request(
                method="get",
                url=self.fav_url,
                params=params,
                cookies=self.cookies,
                wbi=False,
            )
        except Exception as e:
            raise Exception("Error: " + str(e))

        res_json = res.json()

        code = res_json.get("code", -400)

        if code != 0:
            raise ValueError("Error: " + res_json["message"])

        data = res_json["data"]
        for item in data:
            if item["type"] == 2:
                self.id_list.append(item["bvid"])



import json
import os
import random
import re
import time
import urllib.parse
from enum import Enum
from functools import reduce
from hashlib import md5
from typing import Callable

import requests
import typer
from typing_extensions import Annotated

HEADERS = {
    "authority": "api.bilibili.com",
    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "accept-language": "zh-CN,zh;q=0.9",
    "cache-control": "no-cache",
    "dnt": "1",
    "pragma": "no-cache",
    "sec-ch-ua": '"Not A(Brand";v="99", "Google Chrome";v="121", "Chromium";v="121"',
    "sec-ch-ua-mobile": "?0",
    "sec-ch-ua-platform": '"macOS"',
    "sec-fetch-dest": "document",
    "sec-fetch-mode": "navigate",
    "sec-fetch-site": "none",
    "sec-fetch-user": "?1",
    "upgrade-insecure-requests": "1",
    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36",
    "Referer": "https://www.bilibili.com/",
}

mixinKeyEncTab = [
    46,
    47,
    18,
    2,
    53,
    8,
    23,
    32,
    15,
    50,
    10,
    31,
    58,
    3,
    45,
    35,
    27,
    43,
    5,
    49,
    33,
    9,
    42,
    19,
    29,
    28,
    14,
    39,
    12,
    38,
    41,
    13,
    37,
    48,
    7,
    16,
    24,
    55,
    40,
    61,
    26,
    17,
    0,
    1,
    60,
    51,
    30,
    4,
    22,
    25,
    54,
    21,
    56,
    59,
    6,
    63,
    57,
    62,
    11,
    36,
    20,
    34,
    44,
    52,
]


# 音质映射到具体值
audio_quality_map = {
    "64": 30216,
    "132": 30232,
    "192": 30280,
}


class AudioQualityEnums(str, Enum):
    k64 = "64"
    k132 = "132"
    k192 = "192"

    @property
    def quality_id(self):
        return audio_quality_map[self._value_]


def getMixinKey(orig: str):
    "对 imgKey 和 subKey 进行字符顺序打乱编码"
    return reduce(lambda s, i: s + orig[i], mixinKeyEncTab, "")[:32]


def encWbi(params: dict, img_key: str, sub_key: str):
    "为请求参数进行 wbi 签名"
    mixin_key = getMixinKey(img_key + sub_key)
    curr_time = round(time.time())
    params["wts"] = curr_time  # 添加 wts 字段
    params = dict(sorted(params.items()))  # 按照 key 重排参数
    # 过滤 value 中的 "!'()*" 字符
    params = {
        k: "".join(filter(lambda chr: chr not in "!'()*", str(v)))
        for k, v in params.items()
    }
    query = urllib.parse.urlencode(params)  # 序列化参数
    wbi_sign = md5((query + mixin_key).encode()).hexdigest()  # 计算 w_rid
    params["w_rid"] = wbi_sign
    return params


def getWbiKeys() -> tuple[str, str]:
    "获取最新的 img_key 和 sub_key"
    resp = requests.get("https://api.bilibili.com/x/web-interface/nav", headers=HEADERS)
    resp.raise_for_status()
    json_content = resp.json()
    img_url: str = json_content["data"]["wbi_img"]["img_url"]
    sub_url: str = json_content["data"]["wbi_img"]["sub_url"]
    img_key = img_url.rsplit("/", 1)[1].split(".")[0]
    sub_key = sub_url.rsplit("/", 1)[1].split(".")[0]
    return img_key, sub_key


def get_signed_params(params: dict):
    img_key, sub_key = getWbiKeys()
    return encWbi(params, img_key, sub_key)


def gen_dm_args(params: dict):
    """reference: https://github.com/SocialSisterYi/bilibili-API-collect/issues/868"""

    dm_rand = "ABCDEFGHIJK"
    dm_img_list = "[]"
    dm_img_str = "".join(random.sample(dm_rand, 2))
    dm_cover_img_str = "".join(random.sample(dm_rand, 2))
    dm_img_inter = '{"ds":[],"wh":[0,0,0],"of":[0,0,0]}'

    params.update(
        {
            "dm_img_list": dm_img_list,
            "dm_img_str": dm_img_str,
            "dm_cover_img_str": dm_cover_img_str,
            "dm_img_inter": dm_img_inter,
        }
    )

    return params


def get_w_webid(uid: str):
    """reference: https://github.com/SocialSisterYi/bilibili-API-collect/issues/1107#issuecomment-2424269364"""

    dynamic_url = f"https://space.bilibili.com/{uid}/dynamic"

    # TODO: 简化所需的 headers 设置
    headers = {
        "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
        "accept-language": "zh-CN,zh;q=0.9",
        "cache-control": "no-cache",
        "dnt": "1",
        "pragma": "no-cache",
        "priority": "u=0, i",
        "sec-ch-ua": '"Chromium";v="130", "Google Chrome";v="130", "Not?A_Brand";v="99"',
        "sec-ch-ua-mobile": "?0",
        "sec-ch-ua-platform": '"macOS"',
        "sec-fetch-dest": "document",
        "sec-fetch-mode": "navigate",
        "sec-fetch-site": "none",
        "sec-fetch-user": "?1",
        "upgrade-insecure-requests": "1",
        "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/130.0.0.0 Safari/537.36",
    }

    text = requests.get(dynamic_url, headers=headers).text

    # <script id="__RENDER_DATA__" type="application/json">xxx</script>
    __RENDER_DATA__ = re.search(
        r"<script id=\"__RENDER_DATA__\" type=\"application/json\">(.*?)</script>",
        text,
        re.S,
    ).group(1)

    access_id = json.loads(urllib.parse.unquote(__RENDER_DATA__))["access_id"]

    return access_id


def request(
    method: str,
    url: str,
    params: dict = None,
    wbi: bool = False,
    dm: bool = False,
    headers: dict = HEADERS,
    **kwargs,
) -> requests.Response:
    if dm and params is not None:
        params = gen_dm_args(params)
    if wbi and params is not None:
        params = get_signed_params(params)

    return requests.request(method, url, params=params, headers=headers, **kwargs)


def change_directory(directory: str):
    if directory:
        if not os.path.exists(directory):
            os.makedirs(directory)

        if not os.path.isdir(directory):
            raise typer.BadParameter(f"{directory} is not a directory")

        os.chdir(directory)


def check_path(path: str):
    if not os.path.exists(path):
        raise typer.BadParameter(f"{path} does not exist")
    return path


Directory = Annotated[str, typer.Option("-o", "--directory", callback=change_directory)]
Path = typer.Argument(callback=check_path)


class Retry:
    """Retry decorator"""

    def __init__(self, response_succeed, handle_error_response, total=3) -> None:
        self.total = total
        self.__response_succeed = response_succeed
        self.__handle_error_response = handle_error_response
        pass

    def __call__(self, request_func: Callable) -> Callable:
        def wrapped_request(*args, **kwargs):
            for _ in range(self.total):
                res = request_func(*args, **kwargs)
                if self.__response_succeed(res):
                    return res
            self.__handle_error_response(res)
            return None

        return wrapped_request


AudioQuality = Annotated[
    AudioQualityEnums,
    typer.Option("--quality", "-q", help="audio quality", case_sensitive=False),
]



import os
import unittest

from .command import bv, season, series, uid
from .util import AudioQualityEnums, change_directory


class TestBiliFM(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.cpwd = os.getcwd()
        change_directory("tests")

    @classmethod  # 清理测试文件
    def tearDownClass(cls) -> None:
        change_directory(cls.cpwd)
        import shutil

        folder_name = "tests"
        try:
            shutil.rmtree(folder_name)
            print(f"文件夹 {folder_name} 已被删除。")
        except FileNotFoundError:
            print(f"文件夹 {folder_name} 不存在。")
        except Exception as e:
            print(f"删除文件夹 {folder_name} 时出错: {e}")

    def setUp(self):
        # 设置测试环境
        self.test_bv = "BV1yt4y1Q7SS"
        self.audioquality = AudioQualityEnums.k64
        self.test_fav = "69361944"  # 暂时没有找到比较好的测试fav
        self.test_cookies_path = ""
        self.media_id = "69361944"
        self.test_series_uid = "1918016728"
        self.test_series_sid = "4049063"
        self.uid = "7590247"
        self.test_season_uid = "48484716"
        self.test_season_sid = "28722"

    def test_bv_command(self):
        self.assertIsNone(bv(self.test_bv, None, self.audioquality))

    def test_uid_command(self):
        self.assertIsNone(uid(self.uid, None, self.audioquality))

    def test_fav_command(self):
        # 需要cookies 感觉不太能测吧
        pass

    def test_season_command(self):
        self.assertIsNone(
            season(self.test_season_uid, self.test_season_sid, None, self.audioquality)
        )

    def test_series_command(self):
        self.assertIsNone(
            series(self.test_series_uid, self.test_series_sid, None, self.audioquality)
        )


if __name__ == "__main__":
    unittest.main()



"""download bilibili video series, 视频列表"""

import typer

from .util import Retry, request


class Series:
    series_url: str = "https://api.bilibili.com/x/series/archives"
    retry: int = 3

    def __init__(self, uid: str, series_id: str, page_size=30) -> None:
        self.uid = uid
        self.series_id = series_id
        self.page_size = page_size
        self.videos = []
        self.total = 0

    def get_videos(self):
        """return a generator that contain page_size videos"""
        params = {
            "mid": self.uid,
            "series_id": self.series_id,
            "pn": 1,
            "ps": self.page_size,
            "current_id": self.uid,
        }

        @Retry(self.__response_succeed, self.__handle_error_response)
        def wrapped_request():
            """wrap request with retry"""
            return request(method="get", url=self.series_url, params=params).json()

        res = wrapped_request()
        if res is None:
            return 0

        self.total = res["data"]["page"]["total"]

        def bvid_generator():
            for i in range(1, self.total // self.page_size + 2):
                params["pn"] = i
                res = wrapped_request()
                if res:
                    bvids = [ar["bvid"] for ar in res["data"]["archives"]]
                    # self.videos.extend(bvids)
                    yield bvids
                else:
                    typer.echo(
                        f"skip audios from {(i-1)* self.page_size} to {i * self.page_size}"
                    )

        return bvid_generator()

    def __handle_error_response(self, response):
        try:
            archives = response["data"]["archives"]
        except KeyError:
            archives = 0  # something null not none
        if archives is None:
            typer.echo(f"Error: uid {self.uid} or sid {self.series_id} error.")
        else:
            typer.echo("Error: Unknown problem.")
        typer.echo(f"resp: {response}")

    def __response_succeed(self, response) -> bool:
        try:
            return response["data"]["archives"] is not None
        except KeyError:
            return False
